from dataclasses import (
    dataclass,
)

from packer import (
    OptionalPack,
    Pack,
    TypeDescriptor,
    packable,
)


class LengthPrefixedStr(TypeDescriptor):
    __data_size__: int = 2

    @classmethod
    def pack(cls, val: str) -> bytes:
        return int.to_bytes(len((enc := val.encode())), 2, "little") + enc

    @classmethod
    def unpack(cls, data: bytearray) -> tuple[int, str]:
        str_len = int.from_bytes(data[:2], "little")
        return str_len + 2, data[2 : 2 + str_len].decode()


@packable
@dataclass
class CustomTypesStruct:
    test1: Pack[LengthPrefixedStr] = "test"
    test2: OptionalPack[LengthPrefixedStr] = "test2"


def test_custom_packing() -> bool:
    val = CustomTypesStruct().pack()
    assert val == bytearray(b"\x04\x00test\x05\x00test2")

    val = CustomTypesStruct("test", None).pack()
    assert val == bytearray(b"\x04\x00test")


if __name__ == "__main__":
    test_custom_packing()



import math
import struct
from dataclasses import (
    dataclass,
)

from packer import (
    Float,
    Int32,
    OptionalPack,
    Pack,
    packable,
)


@packable
@dataclass
class TestPack:
    id: Pack[Int32] = 69


@packable
@dataclass
class TestPack2:
    test: Pack[Float] = 6.9


@packable
@dataclass
class PackAPackerXd:
    test1: Pack[TestPack]
    test2: Pack[TestPack2]


def test_packer_packing() -> bool:
    t = PackAPackerXd(TestPack(1), TestPack2(1.0))
    data = t.pack()

    assert len(data) == 8 and data == bytearray(b"\x01\x00\x00\x00\x00\x00\x80?")


if __name__ == "__main__":
    test_packer_packing()



import math
import struct
from dataclasses import (
    dataclass,
)

from packer import (
    Float,
    Int32,
    OptionalPack,
    Pack,
    packable,
)


@packable
@dataclass
class SimpleStruct:
    id: Pack[Int32] = 0
    val: OptionalPack[Float] = None


def test_simple_packing() -> bool:
    t = SimpleStruct()
    t.unpack(bytearray([1, 0, 0, 0]) + struct.pack("f", 1.0))

    assert t.id == 1 and t.val and math.isclose(t.val, 1.0, rel_tol=4e-9)

    t = SimpleStruct()
    t.unpack(bytearray([1, 0, 0, 0] * 1))

    assert t.id == 1 and t.val == None
    assert t.pack() == bytearray([1, 0, 0, 0])


if __name__ == "__main__":
    test_simple_packing()



__all__ = ("Int32", "Int16", "Int8")

import sys

from .base import (
    TypeDescriptor,
)

# TODO: Type options, where it's possible to do the following:
"""
@packable
@dataclass
class Test:
    test: Pack[Int32[False]] = 0
"""
# Where the False denotes whether it's a signed or unsigned int.
# This could be done with metaclasses and overriding __getitem__ 👍👍


class Int32(TypeDescriptor):
    __data_size__: int = 4

    @classmethod
    def pack(cls, val: int) -> bytes:
        return val.to_bytes(cls.__data_size__, sys.byteorder, signed=True)

    @classmethod
    def unpack(cls, data: bytearray) -> tuple[int, int]:
        return cls.__data_size__, int.from_bytes(
            data[: cls.__data_size__], sys.byteorder, signed=True
        )


class Int16(TypeDescriptor):
    __data_size__: int = 2

    @classmethod
    def pack(cls, val: int) -> bytes:
        return val.to_bytes(cls.__data_size__, sys.byteorder, signed=True)

    @classmethod
    def unpack(cls, data: bytearray) -> tuple[int, int]:
        return cls.__data_size__, int.from_bytes(
            data[: cls.__data_size__], sys.byteorder, signed=True
        )


class Int8(TypeDescriptor):
    __data_size__: int = 1

    @classmethod
    def pack(cls, val: int) -> bytes:
        return val.to_bytes(cls.__data_size__, sys.byteorder, signed=True)

    @classmethod
    def unpack(cls, data: bytearray) -> tuple[int, int]:
        return cls.__data_size__, int.from_bytes(
            data[: cls.__data_size__], sys.byteorder, signed=True
        )


class UInt32(TypeDescriptor):
    __data_size__: int = 4

    @classmethod
    def pack(cls, val: int) -> bytes:
        return val.to_bytes(cls.__data_size__, sys.byteorder)

    @classmethod
    def unpack(cls, data: bytearray) -> tuple[int, int]:
        return cls.__data_size__, int.from_bytes(data[: cls.__data_size__], sys.byteorder)


class UInt16(TypeDescriptor):
    __data_size__: int = 2

    @classmethod
    def pack(cls, val: int) -> bytes:
        return val.to_bytes(cls.__data_size__, sys.byteorder)

    @classmethod
    def unpack(cls, data: bytearray) -> tuple[int, int]:
        return cls.__data_size__, int.from_bytes(data[: cls.__data_size__], sys.byteorder)


class UInt8(TypeDescriptor):
    __data_size__: int = 1

    @classmethod
    def pack(cls, val: int) -> bytes:
        return val.to_bytes(cls.__data_size__, sys.byteorder)

    @classmethod
    def unpack(cls, data: bytearray) -> tuple[int, int]:
        return cls.__data_size__, int.from_bytes(data[: cls.__data_size__], sys.byteorder)



__all__ = (
    "TypeDescriptor",
    "Pack",
    "OptionalPack",
)

from typing import (
    Any,
    Generic,
    TypeVar,
)

T = TypeVar("T")


class TypeDescriptor:
    __data_size__ = 0

    @classmethod
    def pack(cls, _: T) -> bytes:
        raise NotImplementedError

    @classmethod
    def unpack(cls, _: bytearray) -> tuple[int, Any]:
        raise NotImplementedError


class Pack(Generic[T]): ...


class OptionalPack(Generic[T]): ...



__all__ = ("Float",)

import struct

from .base import (
    TypeDescriptor,
)


class Float(TypeDescriptor):
    __data_size__: int = 4

    @classmethod
    def pack(cls, val: float) -> bytes:
        return struct.pack("f", val)

    @classmethod
    def unpack(cls, data: bytearray) -> tuple[int, int]:
        return cls.__data_size__, struct.unpack("f", data[: cls.__data_size__])[0]



__all__ = ("packable",)

from typing import (
    Callable,
    Protocol,
    Self,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from ._types import *


class PackableProtocol(Protocol):
    _min_size: int
    _packing_data: list[tuple[str, Type, bool, Callable, Callable]]

    def pack(self) -> bytearray: ...
    def unpack(self, data: bytearray) -> int: ...


class Packer:
    def __new__(cls, *args, **kwargs) -> Self:
        instance = super().__new__(cls)
        if getattr(cls, "_packing_data", None):
            return instance

        # attr name, attr_type, is_optional, packer, unpacker
        cls._packing_data = []
        cls._min_size = 0

        type_hints = get_type_hints(instance).items()
        type_hints_list = list(type_hints)

        hint_count = len(type_hints)

        for i, j in enumerate(type_hints):
            attr, attr_type = j
            origin = get_origin(attr_type)
            inner_type = get_args(attr_type)[0]

            next_attr = type_hints_list[i + 1 if i + 1 < hint_count else i][0]
            next_origin = get_origin(type_hints_list[i + 1 if i + 1 < hint_count else i][1])

            if not origin or origin not in {Pack, OptionalPack}:
                continue

            if not issubclass(inner_type, (TypeDescriptor, Packer)):
                raise ValueError(
                    f"The type corresponding with the attribute '{attr}' is marked as Pack but does not subclass TypeDescriptor/Packer."
                )

            optional = True
            if not origin is OptionalPack:
                optional = False
                data_size = 0
                if issubclass(inner_type, (Packer,)):
                    data_size = inner_type._min_size
                else:
                    data_size = inner_type.__data_size__
                cls._min_size += data_size

            if optional and not next_origin is OptionalPack:
                raise ValueError(
                    f"{attr} ({origin}) is followed by {next_attr} ({next_origin}) which is not optional. "
                )

            packer, unpacker = inner_type.pack, inner_type.unpack
            cls._packing_data.append((attr, inner_type, optional, packer, unpacker))

        return instance

    def pack(self) -> bytearray:
        data = bytearray()

        for attr, attr_type, optional, packer, _ in self._packing_data:
            val = getattr(self, attr, None)

            if val is None and not optional:
                raise ValueError(
                    f"The value of {attr} cannot be None as it's not optional ({attr_type})"
                )

            if val is None and optional:  # just break cuz... laikeee???
                # TODO: Debug print to inform the user
                # that we broke (my money up doe on skibidi)
                # out of the loop and will no longer pack the remaining attrs.
                break

            data += packer(val)

        return data

    def unpack(self, data: bytearray) -> int:
        offset = 0
        data_len = len(data)

        if data_len < self._min_size:
            raise ValueError(
                f"Got data of length {data_len} and expected data of length >= {self._min_size}"
            )

        for attr, _, optional, _, unpacker in self._packing_data:
            if offset >= data_len and optional:
                break

            size, val = unpacker(data[offset:])
            offset += size

            setattr(self, attr, val)

        return offset


T = TypeVar("T")


def packable(
    cls: Type[T],
) -> Union[Type[T], Type[PackableProtocol]]:  # I mean... got better ideas??
    class ExtendedCls(cls, Packer): ...

    return ExtendedCls



import math
import struct
from dataclasses import (
    dataclass,
)

from packer import (
    SizedData,
    Pack,
    packable,
)


@packable
@dataclass
class SizedDataStruct:
    sized_data: Pack[SizedData[10]] = None
    sized_data2: Pack[SizedData[50]] = None

def test_sized_packing() -> bool:
    t = SizedDataStruct()
    t.sized_data = bytearray(b"h"*5 + b"i"*5)
    t.sized_data2 = bytearray(50)

    assert len(t.pack()) == 60 and t.pack() == bytearray(b"hhhhhiiiii\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")
    assert t.unpack(bytearray(b"x"*5 + b"d"*5) + bytearray(50)) == 60
    assert t.sized_data == bytearray(b"x"*5 + b"d"*5)
    assert t.sized_data2 == bytearray(50)

if __name__ == "__main__":
    test_sized_packing()



from os import path, walk

from setuptools import setup

setup(
    packages=[
        path.join(root).replace("\\", ".")
        for root, _, files in walk("packer")
        if "__init__.py" in files
    ],
)



__all__ = ("SizedData",)

from typing import Type, Callable
from .base import TypeDescriptor

def _pack_sized_data_wrapper(size: int) -> Callable:
    def inner_packer(val: bytes) -> bytes:
        if (data_len:=len(val)) < size:
            raise ValueError(f"Attempted to pack data of size {data_len}, expected data of size {size}.")
        return val
    return inner_packer
    
def _unpack_sized_data_wrapper(size: int) -> Callable:
    def inner_unpack(data: bytearray) -> tuple[int, bytearray]:
        return (size, data[:size])
    return inner_unpack

class SizedDataMeta(type):
    def __getitem__(cls, data_size: int) -> Type["SizedData"]:
        return type(
            f"{cls.__name__}({data_size})", 
            (TypeDescriptor,), 
            {
                "__data_size__": data_size,
                "pack": _pack_sized_data_wrapper(data_size),
                "unpack": _unpack_sized_data_wrapper(data_size),
            }
        )

class SizedData(metaclass=SizedDataMeta):
    def pack(val: bytes) -> bytes: ...
    def unpack(data: bytearray) -> tuple[int, bytearray]: ...